"""FakePopen class declaration"""
import asyncio
import collections.abc
import io
import os
import signal
import subprocess
import sys
import time
from functools import partial
from typing import Any as AnyType
from typing import Callable
from typing import Dict
from typing import IO
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from . import exceptions
from .types import BUFFER
from .types import OPTIONAL_TEXT
from .types import OPTIONAL_TEXT_OR_ITERABLE
from .utils import Thread


if sys.platform.startswith("win") and sys.version_info < (3, 8):
    COMMAND_SEQ = Sequence[Union[str, bytes]]
else:
    COMMAND_SEQ = Sequence[Union[str, bytes, "os.PathLike[str]", "os.PathLike[bytes]"]]


class FakePopen:
    """Base class that fakes the real subprocess.Popen()"""

    stdout: Optional[BUFFER] = None
    stderr: Optional[BUFFER] = None
    returncode: Optional[int] = None
    text_mode: bool = False
    pid: int = 0

    def __init__(
        self,
        command: Union[
            Union[bytes, str],
            COMMAND_SEQ,
        ],
        stdout: OPTIONAL_TEXT_OR_ITERABLE = None,
        stderr: OPTIONAL_TEXT_OR_ITERABLE = None,
        returncode: int = 0,
        wait: Optional[float] = None,
        callback: Optional[Callable] = None,
        callback_kwargs: Optional[Dict[str, AnyType]] = None,
        signal_callback: Optional[Callable] = None,
        stdin_callable: Optional[Callable] = None,
        **_: Dict[str, AnyType],
    ) -> None:
        if (
            not isinstance(command, (str, bytes))
            and sys.platform.startswith("win")
            and sys.version_info < (3, 8)
        ):
            for arg in command:
                if isinstance(arg, os.PathLike):
                    msg = f"argument of type {arg.__class__.__name__!r} is not iterable"
                    raise TypeError(msg)
        self.args = command
        self.__stdout: OPTIONAL_TEXT_OR_ITERABLE = stdout
        self.__stderr: OPTIONAL_TEXT_OR_ITERABLE = stderr
        self.__returncode: Optional[int] = returncode
        self.__wait: Optional[float] = wait
        self.__thread: Optional[Thread] = None
        self.__callback: Optional[Optional[Callable]] = callback
        self.__callback_kwargs: Optional[Dict[str, AnyType]] = callback_kwargs
        self.__signal_callback: Optional[Callable] = signal_callback
        self.__stdin_callable: Optional[Optional[Callable]] = stdin_callable
        self._signals: List[int] = []

    def __enter__(self) -> "FakePopen":
        return self

    def __exit__(self, *args: List, **kwargs: Dict) -> None:
        if self.__thread and self.__thread.exception:
            raise self.__thread.exception

    def communicate(
        self, input: OPTIONAL_TEXT = None, timeout: Optional[float] = None
    ) -> Tuple[AnyType, AnyType]:
        self._handle_stdin(input)
        self._finalize_thread(timeout)

        if isinstance(self.stdout, asyncio.StreamReader) or isinstance(
            self.stderr, asyncio.StreamReader
        ):
            raise exceptions.PluginInternalError

        return (
            self.stdout.getvalue() if self.stdout else None,
            self.stderr.getvalue() if self.stderr else None,
        )

    def _handle_stdin(self, input: OPTIONAL_TEXT) -> None:
        if input and self.__stdin_callable:
            callable_output = self.__stdin_callable(input)
            if isinstance(callable_output, dict):
                self.stdout = self._extend_stream_from_dict(
                    callable_output, "stdout", self.stdout
                )
                self.stderr = self._extend_stream_from_dict(
                    callable_output, "stderr", self.stderr
                )

    def _finalize_thread(self, timeout: Optional[float]) -> None:
        if self.__thread is None:
            return
        self.__thread.join(timeout)
        if self.returncode is None and self.__returncode is not None:
            self.returncode = self.__returncode
        if self.__thread.exception:
            raise self.__thread.exception

    def _extend_stream_from_dict(
        self, dictionary: Dict[str, AnyType], key: str, stream: Optional[BUFFER]
    ) -> Optional[BUFFER]:
        data = dictionary.get(key)
        if data:
            return self._prepare_buffer(input=data, io_base=stream)
        return None

    def poll(self) -> Optional[int]:
        return self.returncode

    def wait(self, timeout: Optional[float] = None) -> int:
        if timeout and self.__wait and timeout < self.__wait:
            self.__wait -= timeout
            raise subprocess.TimeoutExpired(self.args, timeout)
        self._finalize_thread(timeout)
        if self.returncode is None:
            raise exceptions.PluginInternalError
        return self.returncode

    def send_signal(self, sig: int) -> None:
        self._signals.append(sig)
        if self.__signal_callback:
            self.__signal_callback(self, sig)

    def terminate(self) -> None:
        self.send_signal(signal.SIGTERM)

    def kill(self) -> None:
        if sys.platform == "win32":
            self.terminate()
        else:
            self.send_signal(signal.SIGKILL)

    def configure(self, **kwargs: Optional[Dict]) -> None:
        """Setup the FakePopen instance based on a real Popen arguments."""
        self.__universal_newlines = kwargs.get("universal_newlines", None)
        text = kwargs.get("text", None)
        encoding = kwargs.get("encoding", None)
        errors = kwargs.get("errors", None)

        if text and sys.version_info < (3, 7):
            raise TypeError("__init__() got an unexpected keyword argument 'text'")

        self.text_mode = bool(text or self.__universal_newlines or encoding or errors)

        # validation taken from the real subprocess
        if (
            text is not None
            and self.__universal_newlines is not None
            and bool(self.__universal_newlines) != bool(text)
        ):
            raise subprocess.SubprocessError(
                "Cannot disambiguate when both text "
                "and universal_newlines are supplied but "
                "different. Pass one or the other."
            )

        stdout = kwargs.get("stdout")
        if stdout == subprocess.PIPE:
            self.stdout = self._prepare_buffer(self.__stdout)
        elif isinstance(stdout, (io.BufferedWriter, io.TextIOWrapper)):
            self._write_to_buffer(self.__stdout, stdout)
        stderr = kwargs.get("stderr")
        if stderr == subprocess.STDOUT and self.__stderr:
            assert self.stdout is not None
            self.stdout = self._prepare_buffer(self.__stderr, self.stdout)
        elif stderr == subprocess.PIPE:
            self.stderr = self._prepare_buffer(self.__stderr)
        elif isinstance(stderr, (io.BufferedWriter, io.TextIOWrapper)):
            self._write_to_buffer(self.__stderr, stderr)

    def _prepare_buffer(
        self,
        input: OPTIONAL_TEXT_OR_ITERABLE,
        io_base: Optional[BUFFER] = None,
    ) -> Union[io.BytesIO, io.StringIO, asyncio.StreamReader]:
        linesep = self._convert(os.linesep)

        if isinstance(input, (list, tuple)):
            # need to disable mypy, as input and linesep are unions,
            # mypy thinks that the types might be incompatible, but
            # the _convert() function handles that
            input = linesep.join(map(self._convert, input))  # type: ignore

            # Add trailing newline if data is present.
            if input:
                # same reason to disable mypy as above
                input += linesep  # type: ignore

        if isinstance(input, str) and not self.text_mode:
            input = input.encode()

        if isinstance(input, bytes) and self.text_mode:
            input = input.decode()

        if input and self.__universal_newlines and isinstance(input, str):
            input = input.replace("\r\n", "\n")

        if (
            io_base
            and not isinstance(io_base, asyncio.StreamReader)
            and io_base.tell() == 0
        ):
            # same reason for disabling mypy as in `input = linesep.join...`:
            # both are union so could be incompatible if not _convert()
            input = io_base.getvalue() + (input)  # type: ignore

        if io_base is None:
            io_base = self._get_empty_buffer(self.text_mode)

        if input is None:
            return io_base

        # similar as above - mypy has to be disabled because unions
        if isinstance(io_base, asyncio.StreamReader):
            io_base.feed_data(self._data_to_bytes(input))
        else:
            io_base.write(input)  # type: ignore
        return io_base

    def _get_empty_buffer(self, text: bool) -> BUFFER:
        return io.StringIO() if text else io.BytesIO()

    def _to_bytes(self, data: Sequence[Union[str, bytes]]) -> Sequence[bytes]:
        return [elem if isinstance(elem, bytes) else elem.encode() for elem in data]

    def _data_to_bytes(self, data: OPTIONAL_TEXT_OR_ITERABLE) -> bytes:
        if isinstance(data, collections.abc.Sequence) and not isinstance(data, bytes):
            return b"\n".join(
                (item if isinstance(item, bytes) else item.encode() for item in data)
            )
        if isinstance(data, str):
            return data.encode()
        if not data:
            return b""
        return data

    def _write_to_buffer(self, data: OPTIONAL_TEXT_OR_ITERABLE, buffer: IO) -> None:
        data_type: Callable = (
            # mypy doesn't seem to recognize `partial` as a function
            partial(bytes, encoding=sys.getfilesystemencoding())  # type: ignore
            if "b" in buffer.mode
            else str
        )
        if isinstance(data, (list, tuple)):
            buffer.writelines([data_type(line + "\n") for line in data])
        else:
            buffer.write(data_type(data))

    def _convert(self, input: Union[str, bytes]) -> Union[str, bytes]:
        if isinstance(input, bytes) and self.text_mode:
            return input.decode()
        if isinstance(input, str) and not self.text_mode:
            return input.encode()
        return input

    def _wait(self, wait_period: float) -> None:
        time.sleep(wait_period)
        if self.returncode is None:
            self._finish_process()

    def run_thread(self) -> None:
        """Run the user-defined callback or wait in a thread."""
        if self.__wait is None and self.__callback is None:
            self._finish_process()
        else:
            if self.__callback:
                self.__thread = Thread(
                    target=self.__callback,
                    args=(self,),
                    kwargs=self.__callback_kwargs or {},
                )
            else:
                self.__thread = Thread(target=self._wait, args=(self.__wait,))
            self.__thread.start()

    def _finish_process(self) -> None:
        self.returncode = self.__returncode

        self._finalize_streams()

    def _finalize_streams(self) -> None:
        self._finalize_stream(self.stdout)
        self._finalize_stream(self.stderr)

    def _finalize_stream(self, stream: Optional[BUFFER]) -> None:
        if isinstance(stream, asyncio.StreamReader):
            stream.feed_eof()
        elif stream:
            stream.seek(0)

    def received_signals(self) -> Tuple[int, ...]:
        """Get a tuple of signals received by the process."""
        return tuple(self._signals)


class AsyncFakePopen(FakePopen):
    """Class to handle async processes"""

    stdout: Optional[asyncio.StreamReader]
    stderr: Optional[asyncio.StreamReader]

    async def communicate(  # type: ignore
        self, input: OPTIONAL_TEXT = None, timeout: Optional[float] = None
    ) -> Tuple[AnyType, AnyType]:
        if input:
            # streams were fed with eof, need to be reopened
            await self._reopen_streams()

            self._handle_stdin(input)

            # feed eof one more time as streams were opened
            self._finalize_streams()

        self._finalize_thread(timeout)

        return (
            await self.stdout.read() if self.stdout else None,
            await self.stderr.read() if self.stderr else None,
        )

    async def wait(self, timeout: Optional[float] = None) -> int:  # type: ignore
        return super().wait(timeout)

    def _get_empty_buffer(self, _: bool) -> asyncio.StreamReader:
        return asyncio.StreamReader()

    async def _reopen_streams(self) -> None:
        self.stdout = await self._reopen_stream(self.stdout)
        self.stderr = await self._reopen_stream(self.stderr)

    async def _reopen_stream(
        self, stream: Optional[asyncio.StreamReader]
    ) -> Optional[asyncio.StreamReader]:
        if stream:
            data = await stream.read()
            fresh_stream = self._get_empty_buffer(False)
            fresh_stream.feed_data(data)
            return fresh_stream
        return None



import asyncio
import io
import os
from typing import Sequence
from typing import Union

from .utils import Any
from .utils import Command

OPTIONAL_TEXT = Union[str, bytes, None]
OPTIONAL_TEXT_OR_ITERABLE = Union[
    str,
    bytes,
    None,
    Sequence[Union[str, bytes]],
]
BUFFER = Union[io.BytesIO, io.StringIO, asyncio.StreamReader]
ARGUMENT = Union[str, Any, os.PathLike]
COMMAND = Union[Sequence[ARGUMENT], str, Command]



"""FakeProcess class declaration"""
from collections import defaultdict
from collections import deque
from typing import Any as AnyType
from typing import Callable
from typing import ClassVar
from typing import DefaultDict
from typing import Deque
from typing import Dict
from typing import List
from typing import Optional
from typing import Type
from typing import Union

from . import exceptions
from .process_dispatcher import ProcessDispatcher
from .process_recorder import ProcessRecorder
from .types import COMMAND
from .types import OPTIONAL_TEXT_OR_ITERABLE
from .utils import Any
from .utils import Command
from .utils import Program


class FakeProcess:
    """Main class responsible for process operations"""

    any: ClassVar[Type[Any]] = Any
    program: ClassVar[Type[Program]] = Program

    def __init__(self) -> None:
        self.definitions: DefaultDict[Command, Deque[Union[Dict, bool]]] = defaultdict(
            deque
        )
        self.calls: Deque[COMMAND] = deque()
        self._allow_unregistered: bool = False
        self._keep_last_process: bool = False

        self.exceptions = exceptions

    def register(
        self,
        command: COMMAND,
        stdout: OPTIONAL_TEXT_OR_ITERABLE = None,
        stderr: OPTIONAL_TEXT_OR_ITERABLE = None,
        returncode: int = 0,
        wait: Optional[float] = None,
        callback: Optional[Callable] = None,
        callback_kwargs: Optional[Dict[str, AnyType]] = None,
        signal_callback: Optional[Callable] = None,
        occurrences: int = 1,
        stdin_callable: Optional[Callable] = None,
    ) -> ProcessRecorder:
        """
        Main method for registering the subprocess instances.

        Args:
            command: register the command that will be faked
            stdout: value of the standard output
            stderr: value of the error output
            returncode: return code of the faked process
            wait: artificially wait for the process to finish
            callback: function that will be executed instead of the process
            callback_kwargs: keyword arguments that will be passed into callback
            occurrences: allow multiple usages of the same command
            stdin_callable: function that will interact with stdin
        """
        if wait is not None and callback is not None:
            raise exceptions.IncorrectProcessDefinition(
                "The 'callback' and 'wait' arguments cannot be used "
                "together. Add sleep() to your callback instead."
            )
        if not isinstance(command, Command):
            command = Command(command)

        recorder = ProcessRecorder()
        self.definitions[command].extend(
            [
                {
                    "command": command,
                    "stdout": stdout,
                    "stderr": stderr,
                    "returncode": returncode,
                    "wait": wait,
                    "callback": callback,
                    "callback_kwargs": callback_kwargs,
                    "signal_callback": signal_callback,
                    "stdin_callable": stdin_callable,
                    "recorder": recorder,
                }
            ]
            * occurrences
        )

        return recorder

    register_subprocess = register

    def pass_command(
        self,
        command: COMMAND,
        occurrences: int = 1,
    ) -> None:
        """
        Allow to use a real subprocess together with faked ones.

        Args:
            command: allow to execute the supplied command
            occurrences: allow multiple usages of the same command
        """

        if not isinstance(command, Command):
            command = Command(command)
        self.definitions[command].extend([True] * occurrences)

    def __enter__(self) -> "FakeProcess":
        ProcessDispatcher.register(self)
        return self

    def __exit__(self, *args: List, **kwargs: Dict) -> None:
        ProcessDispatcher.deregister(self)

    def allow_unregistered(self, allow: bool) -> None:
        """
        Allow / block unregistered processes execution. When allowed, the real
        subprocesses will be called. Blocking will raise the exception.

        Args:
            allow: decide whether the unregistered process shall be allowed
        """
        self._allow_unregistered = allow

    def call_count(self, command: COMMAND) -> int:
        """
        Count how many times a certain command was called. Can be used
        together with `fake_process.any()`.

        Args:
            command: lookup command

        Returns:
            number of times a command was called
        """
        if not isinstance(command, Command):
            command_instance = Command(command)
        return len(tuple(filter(lambda elem: elem == command_instance, self.calls)))

    def keep_last_process(self, keep: bool) -> None:
        """
        Keep last process definition from being removed. That can allow / block
        multiple execution of the same command.

        Args:
            keep: decide whether last process shall be kept
        """
        self._keep_last_process = keep

    @classmethod
    def context(cls) -> "FakeProcess":
        """Return a new FakeProcess instance to use it as a context manager."""
        return cls()



from asyncio.subprocess import DEVNULL
from asyncio.subprocess import PIPE
from asyncio.subprocess import Process
from asyncio.subprocess import STDOUT
from asyncio.subprocess import SubprocessStreamProtocol

_ = (DEVNULL, PIPE, STDOUT, Process, SubprocessStreamProtocol)



import faulthandler
import os
import sys

import pytest

pytest_plugins = "pytester"
faulthandler.enable(file=sys.stderr, all_threads=True)


@pytest.fixture(autouse=True)
def setup():
    os.chdir(os.path.dirname(__file__))



class ProcessNotRegisteredError(Exception):
    """
    Raised when the attempted command wasn't registered before.
    Use `fake_process.allow_unregistered(True)` if you want to use real subprocess.
    """


class IncorrectProcessDefinition(Exception):
    """Raised when the register() has been called with wrong arguments"""


class PluginInternalError(Exception):
    """Raised in case of an internal error in the plugin"""



from typing import Generator

import pytest

from . import FakeProcess


@pytest.fixture
def fp() -> Generator[FakeProcess, None, None]:
    """Fake subprocess calls."""
    with FakeProcess() as process:
        yield process


fake_process = fp



import pytest

from pytest_subprocess.utils import Any
from pytest_subprocess.utils import Command


def check_match(command_instance, command):
    assert command_instance == command
    assert command_instance == tuple(command)
    assert command_instance == " ".join(command)
    return True


def check_not_match(command_instance, command):
    assert command_instance != command
    assert command_instance != tuple(command)
    assert command_instance != " ".join(command)
    return True


@pytest.mark.parametrize("command", ("whoami", ["whoami"]))
def test_simple_command(command):
    command = Command(command)
    assert check_match(command, ["whoami"])


@pytest.mark.parametrize("command", ("test command", ("test", "command")))
def test_more_complex_command(command):
    command = Command(command)
    assert check_match(command, ["test", "command"])
    assert check_not_match(command, ["other", "command"])


def test_simple_wildcards():
    command = Command([Any()])
    assert check_match(command, ["test"])
    assert check_match(command, ["not_test"])
    assert check_match(command, ["something", "a", "bit", "longer"])
    assert check_match(command, ["basically", "everything", "will", "match"])

    command = Command(["test", Any()])
    assert check_match(command, ["test", "something"])
    assert check_match(command, ["test", "something_else"])
    assert check_not_match(command, ["something", "test"])
    assert check_match(command, ["test"])

    command = Command([Any(), "test"])
    assert check_match(command, ["something", "test"])
    assert check_match(command, ["something_else", "test"])
    assert check_match(command, ["test"])

    assert check_not_match(command, ["test", "something"])

    command = Command(["test", Any(), "other_test"])
    assert check_match(command, ["test", "something", "other_test"])
    assert check_match(command, ["test", "something_else", "other_test"])
    assert check_match(command, ["test", "other_test"])

    assert check_not_match(command, ["test", "something_else"])


def test_more_complex_wildcards():
    command = Command(["test", Any()])
    assert check_match(command, ["test", "with", "more", "arguments"])

    command = Command(["test", Any(), "end"])
    assert check_match(command, ["test", "blah", "end"])
    assert check_match(command, ["test", "with", "more", "arguments", "end"])
    assert check_not_match(
        command, ["test", "with", "more", "arguments", "invalid_end"]
    )


def test_any_max():
    command = Command([Any(max=1)])
    assert check_match(command, ["test"])
    assert check_match(command, ["other_test"])

    assert check_not_match(command, ["two", "arguments"])
    assert check_not_match(command, ["th", "ree", "arguments"])

    command = Command(["test", Any(max=3)])
    assert check_match(command, ["test", "max", "3", "args"])
    assert check_match(command, ["test", "can_be", "less"])
    assert check_match(command, ["test"])

    assert check_not_match(command, ["wrong", "first", "argument"])
    assert check_not_match(command, ["test", "more", "than", "3", "args"])

    command = Command(["test", Any(max=2), "end"])
    assert check_match(command, ["test", "two", "args", "end"])
    assert check_match(command, ["test", "one_arg", "end"])

    assert check_not_match(command, ["test", "oops", "three", "args", "end"])
    assert check_not_match(command, ["test", "two", "args", "wrong_end"])

    command = Command(["test", Any(max=1), "middle", Any(max=2), "end"])
    assert check_match(
        command, ["test", "one_argument", "middle", "two", "args", "end"]
    )
    assert check_match(command, ["test", "middle", "one_arg", "end"])

    assert check_not_match(
        command, ["test", "two", "args", "middle", "two", "args", "end"]
    )


def test_any_min():
    command = Command([Any(min=2)])
    assert check_match(command, ["any", "two"])
    assert check_match(command, ["or", "even", "three"])

    assert check_not_match(command, ["but_not_one"])

    command = Command(["test", Any(min=1), "end"])
    assert check_match(command, ["test", "with", "more", "arguments", "end"])
    assert check_match(command, ["test", "only_one", "end"])

    assert check_not_match(command, ["only_one", "end"])
    assert check_not_match(command, ["test", "only_one"])
    assert check_not_match(command, ["test", "end"])

    command = Command(["test", Any(min=1), "middle", Any(min=2), "end"])
    assert check_match(
        command, ["test", "one_argument", "middle", "two", "args", "end"]
    )

    assert check_not_match(command, ["test", "middle", "two", "args", "end"])
    assert check_not_match(
        command, ["test", "one_argument", "middle", "one_argument", "end"]
    )


def test_min_max_combined():
    command = Command([Any(min=2, max=4)])
    assert check_match(command, ["just", "two"])
    assert check_match(command, ["now", "three", "arguments"])
    assert check_match(command, ["up", "to", "even", "four"])

    assert check_not_match(command, ["single_argument"])
    assert check_not_match(command, ["five", "is", "little", "too", "many"])

    command = Command(["start", Any(min=1, max=1), "end"])
    assert check_match(command, ["start", "anything", "end"])

    assert check_not_match(command, ["start", "end"])
    assert check_not_match(command, ["start", "too", "many", "end"])


def test_invalid_instantiation():
    with pytest.raises(AttributeError, match="min cannot be greater than max"):
        Any(min=3, max=2)

    with pytest.raises(
        AttributeError, match=r"Cannot use `Any\(\)` one after another."
    ):
        Command([Any(), Any()])

    with pytest.raises(
        TypeError, match="Command can be only of type string, list or tuple."
    ):
        Command(dict(command="ls"))


def test_str_conversions():
    no_arguments = Any()
    assert str(no_arguments) == "Any (min=None, max=None)"

    min_max = Any(min=1, max=3)
    assert str(min_max) == "Any (min=1, max=3)"

    simple_command = Command(["ls", "-lah"])
    assert str(simple_command) == "('ls', '-lah')"

    command_with_any = Command(["ls", Any()])
    assert str(command_with_any) == "('ls', Any (min=None, max=None))"


def test_command_iter():
    """Make sure Command supports iteration"""
    command = Command(["a", "a", "a"])
    assert all(elem == "a" for elem in command)



import asyncio
import os
import sys
import time

import anyio
import pytest

from pytest_subprocess.fake_popen import AsyncFakePopen

PYTHON = sys.executable


@pytest.fixture(autouse=True)
def event_loop(request):
    if sys.platform.startswith("win"):
        if request.node.name.startswith("test_invalid_event_loop"):
            loop = asyncio.SelectorEventLoop()
        else:
            loop = asyncio.ProactorEventLoop()
    elif sys.version_info.minor < 7:
        loop = asyncio.get_event_loop()
    else:
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
    yield loop
    loop.close()


@pytest.mark.asyncio
@pytest.mark.parametrize("mode", ["shell", "exec"])
async def test_basic_usage(fp, mode):
    shell = mode == "shell"
    fp.register(["some-command-that-is-definitely-unavailable"], returncode=500)
    method = (
        asyncio.create_subprocess_shell if shell else asyncio.create_subprocess_exec
    )
    process = await method("some-command-that-is-definitely-unavailable")
    returncode = await process.wait()

    assert process.returncode == returncode
    assert process.returncode == 500


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [True, False])
async def test_with_arguments_shell(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )

    process = await asyncio.create_subprocess_shell(
        f"{PYTHON} example_script.py stderr",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    out, err = await process.communicate()

    assert err == os.linesep.encode().join([b"Stderr line 1", b""])
    assert out == os.linesep.encode().join([b"Stdout line 1", b"Stdout line 2", b""])
    assert process.returncode == 0


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [True, False])
async def test_with_arguments_exec(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )

    process = await asyncio.create_subprocess_exec(
        PYTHON,
        "example_script.py",
        "stderr",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    out, err = await process.communicate()

    assert err == os.linesep.encode().join([b"Stderr line 1", b""])
    assert out == os.linesep.encode().join([b"Stdout line 1", b"Stdout line 2", b""])
    assert process.returncode == 0


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [True, False])
@pytest.mark.parametrize("mode", ["shell", "exec"])
async def test_incorrect_call(fp, fake, mode):
    """Asyncio doesn't support command as a list"""
    shell = mode == "shell"
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(["test"])

    method = (
        asyncio.create_subprocess_shell if shell else asyncio.create_subprocess_exec
    )

    name = "cmd" if shell else "program"
    with pytest.raises(ValueError, match=f"{name} must be a string"):
        await method(["test"])


@pytest.mark.asyncio
@pytest.mark.skipif('sys.platform!="win32"')
@pytest.mark.parametrize("fake", [True, False])
@pytest.mark.parametrize("mode", ["shell", "exec"])
async def test_invalid_event_loop(fp, fake, mode):
    """
    The event_loop is changed by the `event_loop` fixture based on
    the test name (hack).
    """
    shell = mode == "shell"
    fp.allow_unregistered(not fake)
    if fake:
        fp.register([PYTHON, "example_script.py"])

    with pytest.raises(NotImplementedError):
        if shell:
            await asyncio.create_subprocess_shell(f"{PYTHON} example_script.py")
        else:
            await asyncio.create_subprocess_exec(PYTHON, "example_script.py")


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [False, True])
@pytest.mark.parametrize("mode", ["shell", "exec"])
async def test_wait(fp, fake, mode):
    """
    Check that wait argument still works. Unfortunately asyncio doesn't have
    the timeout functionality.
    """
    shell = mode == "shell"

    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "wait", "stderr"],
            stdout="Stdout line 1\nStdout line 2",
            stderr="Stderr line 1",
            wait=0.5,
        )
    method = (
        asyncio.create_subprocess_shell if shell else asyncio.create_subprocess_exec
    )

    command = f"{PYTHON} example_script.py wait stderr"
    if not shell:
        command = command.split()
    else:
        command = [command]

    process = await method(
        *command,
        cwd=os.path.dirname(__file__),
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    assert process.returncode is None

    start_time = time.time()
    returncode = await process.wait()

    assert time.time() - start_time >= 0.45
    assert returncode == 0


@pytest.mark.asyncio
async def test_devnull_stdout(fp):
    """From GitHub #63 - make sure all the `asyncio.subprocess` consts are available."""
    fp.register("cat")

    await asyncio.create_subprocess_exec(
        "cat",
        stdin=asyncio.subprocess.DEVNULL,
        stdout=asyncio.subprocess.STDOUT,
        stderr=asyncio.subprocess.PIPE,
    )


@pytest.mark.asyncio
async def test_anyio(fp):
    await anyio.sleep(0.01)


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [False, True])
async def test_stdout_and_stderr(fp, fake):
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )
    else:
        fp.allow_unregistered(True)

    process = await asyncio.create_subprocess_exec(
        PYTHON,
        "example_script.py",
        "stderr",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    stdout_list = []
    stderr_list = []

    loop = asyncio.get_event_loop()
    await asyncio.gather(
        loop.create_task(_read_stream(process.stdout, stdout_list)),
        loop.create_task(_read_stream(process.stderr, stderr_list)),
        loop.create_task(process.wait()),
    )

    assert stdout_list == [f"Stdout line 1{os.linesep}", f"Stdout line 2{os.linesep}"]
    assert stderr_list == [f"Stderr line 1{os.linesep}"]


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [False, True])
async def test_combined_stdout_and_stderr(fp, fake):
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )
    else:
        fp.allow_unregistered(True)

    process = await asyncio.create_subprocess_exec(
        PYTHON,
        "example_script.py",
        "stderr",
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.STDOUT,
    )

    stdout_list = []
    stderr_list = []

    loop = asyncio.get_event_loop()
    await asyncio.gather(
        loop.create_task(_read_stream(process.stdout, stdout_list)),
        loop.create_task(_read_stream(process.stderr, stderr_list)),
        loop.create_task(process.wait()),
    )

    # sorted() is necessary here, as the order here may be not deterministic,
    # and sometimes stderr comes before stdout or the opposite
    assert sorted(stdout_list) == [
        f"Stderr line 1{os.linesep}",
        f"Stdout line 1{os.linesep}",
        f"Stdout line 2{os.linesep}",
    ]
    assert stderr_list == ["empty"]


async def _read_stream(stream: asyncio.StreamReader, output_list):
    if stream is None:
        output_list.append("empty")
        return None

    while True:
        line = await stream.readline()
        if not line:
            break
        else:
            output_list.append(line.decode())


@pytest.mark.asyncio
@pytest.mark.parametrize("fake", [False, True])
async def test_input(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:

        def stdin_callable(input):
            return {
                "stdout": "Provide an input: Provided: {data}".format(
                    data=input.decode()
                )
            }

        fp.register(
            [PYTHON, "example_script.py", "input"],
            stdout=[b"Stdout line 1", b"Stdout line 2"],
            stdin_callable=stdin_callable,
        )

    process = await asyncio.create_subprocess_exec(
        PYTHON,
        "example_script.py",
        "input",
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
    )
    out, err = await process.communicate(input=b"test")

    assert out.splitlines() == [
        b"Stdout line 1",
        b"Stdout line 2",
        b"Provide an input: Provided: test",
    ]
    assert err is None


@pytest.mark.asyncio
async def test_popen_recorder(fp):
    recorder = fp.register(["test_script"], occurrences=2)
    assert recorder.call_count() == 0

    await asyncio.create_subprocess_exec("test_script")
    assert recorder.call_count() == 1
    await asyncio.create_subprocess_shell("test_script")
    assert recorder.call_count() == 2

    assert all(isinstance(instance, AsyncFakePopen) for instance in recorder.calls)


@pytest.fixture(autouse=True)
def skip_on_pypy():
    """Async test for some reason crash on pypy 3.6 on Windows"""
    if sys.platform == "win32" and sys.version.startswith("3.6"):
        try:
            import __pypy__

            _ = __pypy__
            pytest.skip()
        except ImportError:
            pass



import contextlib
import getpass
import os
import platform
import signal
import subprocess
import sys
import time
from pathlib import Path

import pytest

import pytest_subprocess
from pytest_subprocess.fake_popen import FakePopen

PYTHON = sys.executable

path_or_str = pytest.mark.parametrize(
    "rtype,ptype",
    [
        pytest.param(str, str, id="str"),
        pytest.param(Path, str, id="path,str"),
        pytest.param(str, Path, id="str,path"),
        pytest.param(Path, Path, id="path"),
    ],
)


def setup_fake_popen(monkeypatch):
    """Set the real Popen to a dummy function that just returns input arguments."""
    monkeypatch.setattr(
        pytest_subprocess.process_dispatcher.ProcessDispatcher,
        "built_in_popen",
        lambda command, *args, **kwargs: (command, args, kwargs),
    )


def test_legacy_usage(fake_process):
    cmd = ["cmd"]
    fake_process.register_subprocess(cmd)

    proc = subprocess.run(cmd, check=True)

    assert proc.args == cmd
    assert isinstance(proc.args, type(cmd))


@pytest.mark.parametrize("cmd", [("cmd",), ["cmd"]])
def test_completedprocess_args(fp, cmd):
    fp.register(cmd)

    proc = subprocess.run(cmd, check=True)

    assert proc.args == cmd
    assert isinstance(proc.args, type(cmd))


@path_or_str
def test_completedprocess_args_path(fp, rtype, ptype):
    fp.register([rtype("cmd")])

    if sys.platform.startswith("win") and sys.version_info < (3, 8) and ptype is Path:
        condition = pytest.raises(TypeError)

    else:

        @contextlib.contextmanager
        def null_context():
            yield

        condition = null_context()

    with condition:
        proc = subprocess.run([ptype("cmd")], check=True)

        assert proc.args == [ptype("cmd")]
        assert isinstance(proc.args[0], ptype)


@pytest.mark.parametrize("cmd", [("cmd"), ["cmd"]])
def test_called_process_error(fp, cmd):
    fp.register(cmd, returncode=1)

    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        subprocess.run(cmd, check=True)

    assert exc_info.value.cmd == cmd
    assert isinstance(exc_info.value.cmd, type(cmd))


@pytest.mark.parametrize("cmd", [("cmd"), ["cmd"]])
def test_called_process_error_with_any(fp, cmd):
    fp.register([fp.any()], returncode=1)

    with pytest.raises(subprocess.CalledProcessError) as exc_info:
        subprocess.run(cmd, check=True)

    assert exc_info.value.cmd == cmd
    assert isinstance(exc_info.value.cmd, type(cmd))


def test_keep_last_process_error_with_any(fp):
    fp.register([fp.any()], returncode=1)
    fp.keep_last_process(True)

    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(["cmd"], check=True)

    with pytest.raises(subprocess.CalledProcessError):
        subprocess.run(["cmd2"], check=True)


def test_multiple_levels(fp):
    """Register fake process on different levels and check the behavior"""

    # process definition on the top level
    fp.register(("first_command"), stdout="first command top-level")

    with fp.context() as nested:
        # lower level, override the same command and define new one
        nested.register("first_command", stdout="first command lower-level")
        nested.register("second_command", stdout="second command lower-level")

        assert (
            subprocess.check_output("first_command")
            == ("first command lower-level").encode()
        )
        assert (
            subprocess.check_output("second_command")
            == ("second command lower-level").encode()
        )

    # first command definition shall be back at top-level definition, and the second
    # command is no longer defined so it shall raise an exception
    assert (
        subprocess.check_output("first_command") == ("first command top-level").encode()
    )
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("second_command")
    assert str(exc.value) == "The process 'second_command' was not registered."


def test_not_registered(fp, monkeypatch):
    """
    Scenario with attempt of running a command that is not registered.

    First two tries will raise an exception, but the last one will set
    `process.allow_unregistered(True)` which will allow to execute the process.
    """
    assert fp

    # this one will use exception from `pytest_subprocess.ProcessNotRegisteredError`
    # to make sure it still works (for backwards compatibility)
    with pytest.raises(pytest_subprocess.ProcessNotRegisteredError) as exc:
        subprocess.Popen("test")

    assert str(exc.value) == "The process 'test' was not registered."

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.Popen(("test", "with", "args"))

    assert str(exc.value) == "The process 'test with args' was not registered."

    fp.allow_unregistered(True)
    setup_fake_popen(monkeypatch)
    result = subprocess.Popen("test", shell=True)

    assert result == ("test", (), {"shell": True})


def test_context(fp, monkeypatch):
    """Test context manager behavior."""
    setup_fake_popen(monkeypatch)

    with fp.context() as nested:
        nested.register("test")
        subprocess.Popen("test")

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.Popen("test")

    assert str(exc.value) == "The process 'test' was not registered."


@pytest.mark.parametrize("fake", [False, True])
@path_or_str
def test_basic_process(fp, fake, rtype, ptype):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [rtype(PYTHON), "example_script.py"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=None,
        )

    if sys.platform.startswith("win") and sys.version_info < (3, 8) and ptype is Path:
        condition = pytest.raises(TypeError)

    else:

        @contextlib.contextmanager
        def null_context():
            yield

        condition = null_context()

    with condition:
        process = subprocess.Popen(
            [ptype(PYTHON), "example_script.py"],
            cwd=os.path.dirname(__file__),
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )
        out, err = process.communicate()

        assert process.poll() == 0
        assert process.returncode == 0
        assert process.pid > 0

        # splitlines is required to ignore differences between LF and CRLF
        assert out.splitlines() == [b"Stdout line 1", b"Stdout line 2"]
        assert err == b""


@pytest.mark.parametrize("fake", [False, True])
def test_basic_process_merge_streams(fp, fake):
    """Stderr is merged into stdout."""
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "-u", "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )

    process = subprocess.Popen(
        [PYTHON, "-u", "example_script.py", "stderr"],
        cwd=os.path.dirname(__file__),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )
    out, err = process.communicate()

    assert out.splitlines() == [
        b"Stdout line 1",
        b"Stdout line 2",
        b"Stderr line 1",
    ]
    assert err is None


@pytest.mark.parametrize("fake", [False, True])
def test_wait(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "wait", "stderr"],
            stdout="Stdout line 1\nStdout line 2",
            stderr="Stderr line 1",
            wait=0.5,
        )
    process = subprocess.Popen(
        (PYTHON, "example_script.py", "wait", "stderr"),
        cwd=os.path.dirname(__file__),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    )

    assert process.returncode is None

    with pytest.raises(subprocess.TimeoutExpired) as exc:
        process.wait(timeout=0.1)
    assert (
        str(exc.value).replace("\\\\", "\\")
        == f"Command '('{PYTHON}', 'example_script.py', 'wait', 'stderr')' "
        "timed out after 0.1 seconds"
    )

    assert process.wait() == 0


@pytest.mark.parametrize("fake", [False, True])
def test_check_output(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout="Stdout line 1\nStdout line 2",
        )
    process = subprocess.check_output((PYTHON, "example_script.py"))

    assert process.splitlines() == [b"Stdout line 1", b"Stdout line 2"]


@pytest.mark.parametrize("fake", [False, True])
def test_check_call(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout="Stdout line 1\nStdout line 2\n",
        )
        fp.register([PYTHON, "example_script.py", "non-zero"], returncode=1)
    assert subprocess.check_call((PYTHON, "example_script.py")) == 0

    # check also non-zero exit code
    with pytest.raises(subprocess.CalledProcessError) as exc:
        assert subprocess.check_call((PYTHON, "example_script.py", "non-zero")) == 1

    if sys.version_info >= (3, 6):
        assert (
            str(exc.value).replace("\\\\", "\\")
            == f"Command '('{PYTHON}', 'example_script.py', 'non-zero')' "
            "returned non-zero exit status 1."
        )


@pytest.mark.parametrize("fake", [False, True])
def test_call(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout="Stdout line 1\nStdout line 2\n",
        )
    assert subprocess.call((PYTHON, "example_script.py")) == 0


@pytest.mark.parametrize("fake", [False, True])
@pytest.mark.skipif(
    sys.version_info <= (3, 5),
    reason="subprocess.run() was introduced in python3.4",
)
def test_run(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout=["Stdout line 1", "Stdout line 2"],
        )
    process = subprocess.run((PYTHON, "example_script.py"), stdout=subprocess.PIPE)

    assert process.returncode == 0
    assert process.stdout == os.linesep.encode().join(
        [b"Stdout line 1", b"Stdout line 2", b""]
    )
    assert process.stderr is None


@pytest.mark.parametrize("fake", [False, True])
def test_universal_newlines(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout=b"Stdout line 1\r\nStdout line 2\r\n",
        )
    process = subprocess.Popen(
        (PYTHON, "example_script.py"), universal_newlines=True, stdout=subprocess.PIPE
    )
    process.wait()

    assert process.stdout.read() == "Stdout line 1\nStdout line 2\n"


@pytest.mark.parametrize("fake", [False, True])
def test_text(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py"],
            stdout=[b"Stdout line 1", b"Stdout line 2"],
        )
    if sys.version_info < (3, 7):
        with pytest.raises(TypeError) as exc:
            subprocess.Popen(
                (PYTHON, "example_script.py"), stdout=subprocess.PIPE, text=True
            )
        assert str(exc.value) == "__init__() got an unexpected keyword argument 'text'"
    else:
        process = subprocess.Popen(
            (PYTHON, "example_script.py"), stdout=subprocess.PIPE, text=True
        )
        process.wait()

        assert process.stdout.read().splitlines() == ["Stdout line 1", "Stdout line 2"]


def test_binary(fp):
    fp.register(
        ["some-cmd"],
        stdout=bytes.fromhex("aabbcc"),
    )

    process = subprocess.Popen(["some-cmd"], stdout=subprocess.PIPE)
    process.wait()

    assert process.stdout.read() == b"\xaa\xbb\xcc"


def test_empty_stdout(fp):
    fp.register(["some-cmd"], stdout=b"")

    process = subprocess.Popen(["some-cmd"], stdout=subprocess.PIPE)
    process.wait()

    assert process.stdout.read() == b""


def test_empty_stdout_list(fp):
    fp.register(["some-cmd"], stdout=[])

    process = subprocess.Popen(["some-cmd"], stdout=subprocess.PIPE)
    process.wait()

    assert process.stdout.read() == b""


@pytest.mark.parametrize("fake", [False, True])
def test_input(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:

        def stdin_callable(input):
            return {
                "stdout": "Provide an input: Provided: {data}".format(
                    data=input.decode()
                )
            }

        fp.register(
            [PYTHON, "example_script.py", "input"],
            stdout=[b"Stdout line 1", b"Stdout line 2"],
            stdin_callable=stdin_callable,
        )

    process = subprocess.Popen(
        (PYTHON, "example_script.py", "input"),
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
    )
    out, err = process.communicate(input=b"test")

    assert out.splitlines() == [
        b"Stdout line 1",
        b"Stdout line 2",
        b"Provide an input: Provided: test",
    ]
    assert err is None


@pytest.mark.skipif(
    sys.version_info < (3, 7),
    reason="No need to test since 'text' is available since 3.7",
)
@pytest.mark.parametrize("fake", [False, True])
def test_ambiguous_input(fp, fake):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register("test", occurrences=2)

    with pytest.raises(subprocess.SubprocessError) as exc:
        subprocess.run("test", universal_newlines=False, text=True)

    assert str(exc.value) == (
        "Cannot disambiguate when both text "
        "and universal_newlines are supplied but "
        "different. Pass one or the other."
    )

    with pytest.raises(subprocess.SubprocessError) as exc:
        subprocess.run("test", universal_newlines=True, text=False)

    assert str(exc.value) == (
        "Cannot disambiguate when both text "
        "and universal_newlines are supplied but "
        "different. Pass one or the other."
    )


@pytest.mark.flaky(reruns=2, condition=platform.python_implementation() == "PyPy")
@pytest.mark.parametrize("fake", [False, True])
def test_multiple_wait(fp, fake):
    """
    Wait multiple times for 0.2 seconds with process lasting for 0.5.
    Third wait shall is a bit longer and will not raise an exception,
    due to exceeding the subprocess runtime.
    """
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "example_script.py", "wait"],
            wait=0.5,
        )

    process = subprocess.Popen(
        (PYTHON, "example_script.py", "wait"),
    )
    with pytest.raises(subprocess.TimeoutExpired):
        process.wait(timeout=0.2)

    with pytest.raises(subprocess.TimeoutExpired):
        process.wait(timeout=0.1)

    process.wait(0.4)

    assert process.returncode == 0

    # one more wait shall do no harm
    process.wait(0.2)


def test_wrong_arguments(fp):
    with pytest.raises(fp.exceptions.IncorrectProcessDefinition) as exc:
        fp.register("command", wait=1, callback=lambda _: True)

    assert str(exc.value) == (
        "The 'callback' and 'wait' arguments cannot be used "
        "together. Add sleep() to your callback instead."
    )


def test_callback(fp, capsys):
    """
    This test will show a usage of the callback argument.
    The callback argument will have access to the FakePopen so it will
    change the returncode.
    One callback execution will also pass a keyword argument.
    """

    def callback(process, argument=None):
        print("from callback with argument={}".format(argument))
        process.returncode = 1

    fp.register("test", callback=callback)
    fp.register("test", callback=callback, callback_kwargs={"argument": "value"})

    assert subprocess.call("test") == 1
    assert capsys.readouterr().out == "from callback with argument=None\n"

    assert subprocess.call("test") == 1
    assert capsys.readouterr().out == "from callback with argument=value\n"


def test_mutiple_occurrences(fp):
    # register 3 occurrences of the same command at once
    fp.register("test", occurrences=3)

    process_1 = subprocess.Popen("test")
    assert process_1.returncode == 0
    process_2 = subprocess.Popen("test")
    assert process_2.returncode == 0
    assert process_2.pid == process_1.pid + 1
    process_3 = subprocess.Popen("test")
    assert process_3.returncode == 0
    assert process_3.pid == process_2.pid + 1
    # 4-th time shall raise an exception
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("test")

    assert str(exc.value) == "The process 'test' was not registered."


def test_different_output(fp):
    # register process with output changing each execution
    fp.register("test", stdout="first execution")
    fp.register("test", stdout="second execution")
    # the third execution will return non-zero exit code
    fp.register("test", stdout="third execution", returncode=1)

    assert subprocess.check_output("test") == b"first execution"
    assert subprocess.check_output("test") == b"second execution"
    third_process = subprocess.Popen("test", stdout=subprocess.PIPE)
    assert third_process.stdout.read() == b"third execution"
    assert third_process.returncode == 1

    # 4-th time shall raise an exception
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("test")

    assert str(exc.value) == "The process 'test' was not registered."

    # now, register two processes once again, but the last one will be kept forever
    fp.register("test", stdout="first execution")
    fp.register("test", stdout="second execution")
    fp.keep_last_process(True)

    # now the processes can be called forever
    assert subprocess.check_output("test") == b"first execution"
    assert subprocess.check_output("test") == b"second execution"
    assert subprocess.check_output("test") == b"second execution"
    assert subprocess.check_output("test") == b"second execution"


def test_different_output_with_context(fp):
    """
    Leaving one context shall bring back the upper contexts processes
    even if they were already consumed. This functionality is important
    to allow a broader-level fixtures that register own processes and keep
    them predictable.
    """
    fp.register("test", stdout="top-level")

    with fp.context() as nested:
        nested.register("test", stdout="nested")

        assert subprocess.check_output("test") == b"nested"
        assert subprocess.check_output("test") == b"top-level"

        with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
            subprocess.check_call("test")

        assert str(exc.value) == "The process 'test' was not registered."

    with fp.context() as nested2:
        # another nest level, the top level shall reappear
        nested2.register("test", stdout="nested2")

        assert subprocess.check_output("test") == b"nested2"
        assert subprocess.check_output("test") == b"top-level"

        with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
            subprocess.check_call("test")

        assert str(exc.value) == "The process 'test' was not registered."

    assert subprocess.check_output("test") == b"top-level"

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("test")

    assert str(exc.value) == "The process 'test' was not registered."


def test_different_output_with_context_multilevel(fp):
    """
    This is a similar test to the previous one, but here the nesting will be deeper
    """
    fp.register("test", stdout="top-level")

    with fp.context() as first_level:
        first_level.register("test", stdout="first-level")

        with fp.context() as second_level:
            second_level.register("test", stdout="second-level")

            assert subprocess.check_output("test") == b"second-level"
            assert subprocess.check_output("test") == b"first-level"
            assert subprocess.check_output("test") == b"top-level"

            with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
                subprocess.check_call("test")

        assert subprocess.check_output("test") == b"first-level"
        assert subprocess.check_output("test") == b"top-level"

        with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
            subprocess.check_call("test")

        assert str(exc.value) == "The process 'test' was not registered."

    assert subprocess.check_output("test") == b"top-level"

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("test")


def test_multiple_level_early_consuming(fp):
    """
    The top-level will be declared with two ocurrences, but the first one will
    be consumed before entering the context manager.
    """
    fp.register("test", stdout="top-level", occurrences=2)
    assert subprocess.check_output("test") == b"top-level"

    with fp.context():
        assert subprocess.check_output("test") == b"top-level"

        with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
            subprocess.check_call("test")

        assert str(exc.value) == "The process 'test' was not registered."

    assert subprocess.check_output("test") == b"top-level"

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError) as exc:
        subprocess.check_call("test")

    assert str(exc.value) == "The process 'test' was not registered."


def test_keep_last_process(fp):
    """
    The ProcessNotRegisteredError will never be raised for the process that
    has been registered at least once.
    """
    fp.keep_last_process(True)
    fp.register("test", stdout="First run")
    fp.register("test", stdout="Second run")

    assert subprocess.check_output("test") == b"First run"
    assert subprocess.check_output("test") == b"Second run"
    assert subprocess.check_output("test") == b"Second run"
    assert subprocess.check_output("test") == b"Second run"


def test_git(fp):
    fp.register(["git", "branch"], stdout=["* fake_branch", "  master"])

    process = subprocess.Popen(
        ["git", "branch"], stdout=subprocess.PIPE, universal_newlines=True
    )
    out, _ = process.communicate()

    assert process.returncode == 0
    assert out == "* fake_branch\n  master\n"


def test_use_real(fp):
    fp.pass_command([PYTHON, "example_script.py"], occurrences=3)
    fp.register([PYTHON, "example_script.py"], stdout=["Fake line 1", "Fake line 2"])

    for _ in range(0, 3):
        assert (
            subprocess.check_output(
                [PYTHON, "example_script.py"], universal_newlines=True
            )
            == "Stdout line 1\nStdout line 2\n"
        )
    assert (
        subprocess.check_output([PYTHON, "example_script.py"], universal_newlines=True)
        == "Fake line 1\nFake line 2\n"
    )


@pytest.mark.skipif(os.name == "nt", reason="Skip on windows")
def test_real_process(fp):
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        # this will fail, as "ls" command is not registered
        subprocess.call("ls")

    fp.pass_command("ls")
    # now it should be fine
    assert subprocess.call("ls") == 0

    # allow all commands to be called by real subprocess
    fp.allow_unregistered(True)
    assert subprocess.call(["ls", "-l"]) == 0


def test_context_manager(fp):
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        # command not registered, so will raise an exception
        subprocess.check_call("test")

    with fp.context() as nested_process:
        nested_process.register("test", occurrences=3)
        # now, we can call the command 3 times without error
        assert subprocess.check_call("test") == 0
        assert subprocess.check_call("test") == 0

    # the command was called 2 times, so one occurrence left, but since the
    # context manager has been left, it is not registered anymore
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.check_call("test")


def test_raise_exception(fp):
    def callback_function(process):
        process.returncode = 1
        raise PermissionError("exception raised by subprocess")

    fp.register(["test"], callback=callback_function)

    with pytest.raises(PermissionError, match="exception raised by subprocess"):
        process = subprocess.Popen(["test"])
        process.wait()

    assert process.returncode == 1


def test_callback_with_arguments(fp):
    def callback_function(process, return_code):
        process.returncode = return_code

    return_code = 127

    fp.register(
        ["test"],
        callback=callback_function,
        callback_kwargs={"return_code": return_code},
    )

    process = subprocess.Popen(["test"])
    process.wait()

    assert process.returncode == return_code


def test_subprocess_pipe_without_stream_definition(fp):
    """
    From GitHub #17 - the fake_subprocess was crashing if the subprocess was called
    with stderr=subprocess.PIPE but the stderr was not defined during the process
    registration.
    """
    fp.register(
        ["test-no-stderr"],
        stdout="test",
    )
    fp.register(
        ["test-no-stdout"],
        stderr="test",
    )
    fp.register(
        ["test-no-streams"],
    )

    assert (
        subprocess.check_output(["test-no-stderr"], stderr=subprocess.STDOUT).decode()
        == "test"
    )
    assert (
        subprocess.check_output(["test-no-stdout"], stderr=subprocess.STDOUT).decode()
        == "test"
    )
    assert (
        subprocess.check_output(["test-no-streams"], stderr=subprocess.STDOUT).decode()
        == ""
    )


@pytest.mark.parametrize("command", (("test",), "test"))
def test_different_command_type(fp, command):
    """
    From GitHub #18 - registering process as ["command"] or "command" should make no
    difference, and none of those command usage attempts shall raise error.
    """
    fp.keep_last_process(True)

    fp.register(command)

    assert subprocess.check_call("test") == 0
    assert subprocess.check_call(["test"]) == 0


@pytest.mark.parametrize(
    "command", (("test", "with", "arguments"), "test with arguments")
)
def test_different_command_type_complex_command(fp, command):
    """
    Similar to previous test, but the command is more complex.
    """
    fp.keep_last_process(True)

    fp.register(command)

    assert subprocess.check_call("test with arguments") == 0
    assert subprocess.check_call(["test", "with", "arguments"]) == 0


@pytest.mark.flaky(reruns=2, condition=platform.python_implementation() == "PyPy")
def test_raise_exception_check_output(fp):
    """
    From GitHub#16 - the check_output raises the CalledProcessError exception
    when the exit code is not zero. The exception should not shadow the exception
    from the callback, if any.

    For some reason, this test is flaky on PyPy. Further investigation required.
    """

    def callback_function(_):
        raise FileNotFoundError("raised in callback")

    fp.register("regular-behavior", returncode=1)
    fp.register("custom-exception", returncode=1, callback=callback_function)

    with pytest.raises(subprocess.CalledProcessError):
        subprocess.check_output("regular-behavior")

    with pytest.raises(FileNotFoundError, match="raised in callback"):
        subprocess.check_output("custom-exception")


def test_callback_and_return_code(fp):
    """Regression - the returncode was ignored when callback_function was present."""

    def dummy_callback(_):
        pass

    def override_returncode(process):
        process.returncode = 5

    return_code = 1

    fp.register("test-dummy", returncode=return_code, callback=dummy_callback)

    process = subprocess.Popen("test-dummy")
    process.wait()

    assert process.returncode == return_code

    fp.register("test-increment", returncode=return_code, callback=override_returncode)

    process = subprocess.Popen("test-increment")
    process.wait()

    assert process.returncode == 5


@pytest.mark.skipif(
    sys.version_info <= (3, 6),
    reason="encoding and errors has been introduced in 3.6",
)
@pytest.mark.parametrize("argument", ["encoding", "errors"])
@pytest.mark.parametrize("fake", [False, True])
def test_encoding(fp, fake, argument):
    """If encoding or errors is provided, the `text=True` behavior should be enabled."""
    username = getpass.getuser()
    values = {"encoding": "utf-8", "errors": "strict"}

    fp.allow_unregistered(not fake)
    if fake:
        fp.register(["whoami"], stdout=username)

    output = subprocess.check_output(
        ["whoami"], **{argument: values.get(argument)}
    ).strip()

    assert isinstance(output, str)
    assert output.endswith(username)


@pytest.mark.parametrize("command", ["ls -lah", ["ls", "-lah"]])
def test_string_or_tuple(fp, command):
    """
    It doesn't matter how you register the command, it should work as string or list.
    """
    fp.register(command, occurrences=2)
    assert subprocess.check_call("ls -lah") == 0
    assert subprocess.check_call(["ls", "-lah"]) == 0


def test_with_wildcards(fp):
    """Use Any() with real example"""
    fp.keep_last_process(True)
    fp.register(("ls", fp.any()))

    assert subprocess.check_call("ls -lah") == 0
    assert subprocess.check_call(["ls", "-lah", "/tmp"]) == 0
    assert subprocess.check_call(["ls"]) == 0

    fp.register(["cp", fp.any(min=2)])
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.check_call("cp /source/dir")
    assert subprocess.check_call("cp /source/dir /tmp/random-dir") == 0

    fp.register(["cd", fp.any(max=1)])
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.check_call(["cd ~/ /tmp"])
    assert subprocess.check_call("cd ~/") == 0


def test_with_program(fp, monkeypatch):
    """Use Program() with real example"""
    fp.keep_last_process(True)
    fp.register((fp.program("ls"), fp.any()))

    assert subprocess.check_call("ls -lah") == 0
    assert subprocess.check_call(["/ls", "-lah", "/tmp"]) == 0
    assert subprocess.check_call(["/usr/bin/ls"]) == 0

    with monkeypatch.context():
        monkeypatch.setattr(sys, "platform", "win32")
        monkeypatch.setenv("PATHEXT", ".EXE")
        assert subprocess.check_call("ls.EXE -lah") == 0
        assert subprocess.check_call("ls.exe -lah") == 0

    fp.register([fp.program("cp"), fp.any()])
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.check_call(["other"])
    assert subprocess.check_call("cp /source/dir /tmp/random-dir") == 0

    fp.register([fp.program("cd")])
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.check_call(["cq"])
    assert subprocess.check_call("cd") == 0


def test_call_count(fp):
    """Check if commands are registered and counted properly"""
    fp.keep_last_process(True)
    fp.register([fp.any()])

    assert subprocess.check_call("ls -lah") == 0
    assert subprocess.check_call(["cp", "/tmp/source", "/source"]) == 0
    assert subprocess.check_call(["cp", "/source", "/destination"]) == 0
    assert subprocess.check_call(["cp", "/source", "/other/destination"]) == 0

    assert "ls -lah" in fp.calls
    assert ["cp", "/tmp/source", "/source"] in fp.calls
    assert ["cp", "/source", "/destination"] in fp.calls
    assert ["cp", "/source", "/other/destination"] in fp.calls

    assert fp.call_count("cp /tmp/source /source") == 1
    assert fp.call_count(["cp", "/source", fp.any()]) == 2
    assert fp.call_count(["cp", fp.any()]) == 3
    assert fp.call_count(["ls", "-lah"]) == 1


def test_called_process_waits_for_the_callback_to_finish(fp, tmp_path):
    output_file_path = tmp_path / "output"

    def callback(process):
        # simulate a long-running process that creates an output file at the very end
        time.sleep(1)
        output_file_path.touch()

    fp.register([fp.any()], callback=callback)
    subprocess.run(["ls", "-al"], stdin="abc")

    assert output_file_path.exists()


@pytest.mark.parametrize("method", [FakePopen.wait, FakePopen.communicate])
def test_raises_exceptions_from_callback(fp, method):
    """Make sure that both .wait() and .communicate() raise exception from callback"""

    class MyException(Exception):
        pass

    def callback(process):
        raise MyException()

    fp.register(["test"], callback=callback)

    proc = subprocess.Popen("test")
    with pytest.raises(MyException):
        method(proc)


def test_allow_unregistered_cleaning(fp):
    """
    GitHub: #46.
    The `allow_unregistered()` function should affect only the level where it was applied
    The setting shouldn't leak to a higher levels or other tests.
    """
    fp.allow_unregistered(False)

    with fp.context() as context:
        context.allow_unregistered(True)

        subprocess.run([PYTHON, "example_script.py"])
        subprocess.run([PYTHON, "example_script.py"])
        subprocess.run([PYTHON, "example_script.py"])

    with fp.context():
        with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
            subprocess.run([PYTHON, "example_script.py"])

    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.run(["test"])


def test_keep_last_process_cleaning(fp):
    """
    GitHub: #46.
    The `keep_last_process()` function should affect only the level where it was applied
    The setting shouldn't leak to a higher levels or other tests.
    """
    fp.keep_last_process(False)

    with fp.context() as context:
        context.keep_last_process(True)
        context.register(["test"])

        subprocess.run(["test"])
        subprocess.run(["test"])
        subprocess.run(["test"])

    with fp.context():
        with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
            subprocess.run(["test"])

    fp.register(["test"])
    subprocess.run(["test"])
    with pytest.raises(fp.exceptions.ProcessNotRegisteredError):
        subprocess.run(["test"])


def test_signals(fp):
    """Test signal receiving functionality"""
    fp.register("test")

    process = subprocess.Popen("test")

    process.kill()
    process.terminate()
    process.send_signal(signal.SIGSEGV)

    if sys.platform == "win32":
        expected_signals = (signal.SIGTERM, signal.SIGTERM, signal.SIGSEGV)
    else:
        expected_signals = (signal.SIGKILL, signal.SIGTERM, signal.SIGSEGV)

    assert process.received_signals() == expected_signals


def test_signal_callback(fp):
    """Test that signal callbacks work."""

    def callback(process, sig):
        if sig == signal.SIGTERM:
            process.returncode = -1

    fp.register("test", signal_callback=callback, occurrences=3)

    # no signal
    process = subprocess.Popen("test")
    process.wait()

    assert process.returncode == 0

    # other signal
    process = subprocess.Popen("test")
    process.send_signal(signal.SIGSEGV)
    process.wait()

    assert process.returncode == 0

    # sigterm
    process = subprocess.Popen("test")
    process.send_signal(signal.SIGTERM)
    process.wait()

    assert process.returncode == -1


@pytest.mark.parametrize("fake", [False, True])
@pytest.mark.parametrize("bytes", [True, False])
def test_non_piped_streams(tmpdir, fp, fake, bytes):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "-u", "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr=["Stderr line 1"],
        )

    stdout_path = tmpdir.join("stdout")
    stderr_path = tmpdir.join("stderr")

    mode = "wb" if bytes else "w"

    with open(stdout_path, mode) as stdout, open(stderr_path, mode) as stderr:
        process = subprocess.Popen(
            [PYTHON, "-u", "example_script.py", "stderr"],
            stdout=stdout,
            stderr=stderr,
        )

        err, out = process.communicate()

    assert out is None
    assert err is None

    with open(stdout_path, "r") as stdout, open(stderr_path, "r") as stderr:
        out = stdout.readlines()
        err = stderr.readlines()

    assert out == ["Stdout line 1\n", "Stdout line 2\n"]
    assert err == ["Stderr line 1\n"]


@pytest.mark.parametrize("fake", [False, True])
@pytest.mark.parametrize("bytes", [True, False])
def test_non_piped_same_file(tmpdir, fp, fake, bytes):
    fp.allow_unregistered(not fake)
    if fake:
        fp.register(
            [PYTHON, "-u", "example_script.py", "stderr"],
            stdout=["Stdout line 1", "Stdout line 2"],
            stderr="Stderr line 1\n",
        )

    output_path = tmpdir.join("output")

    mode = "wb" if bytes else "w"

    with open(output_path, mode) as out_file:
        process = subprocess.Popen(
            [PYTHON, "-u", "example_script.py", "stderr"],
            stdout=out_file,
            stderr=out_file,
        )

        err, out = process.communicate()

    assert out is None
    assert err is None

    with open(output_path, "r") as out_file:
        output = out_file.readlines()

    assert output == ["Stdout line 1\n", "Stdout line 2\n", "Stderr line 1\n"]


def test_process_recorder(fp):
    fp.keep_last_process(True)
    recorder = fp.register(["test_script", fp.any()])
    assert recorder.calls == []
    assert recorder.call_count() == 0
    assert not recorder.was_called()

    subprocess.call(("test_script", "random_argument"))
    assert recorder.call_count() == 1
    assert recorder.was_called()
    assert recorder.was_called(("test_script", "random_argument"))
    assert not recorder.was_called(("test_script", "another_argument"))

    subprocess.Popen(["test_script", "another_argument"])
    assert recorder.call_count() == 2
    assert recorder.was_called(("test_script", "another_argument"))

    assert recorder.call_count(["test_script", "random_argument"]) == 1
    assert recorder.call_count(["test_script", "another_argument"]) == 1

    assert recorder.first_call.args == ("test_script", "random_argument")
    assert recorder.last_call.args == ["test_script", "another_argument"]

    assert all(isinstance(instance, FakePopen) for instance in recorder.calls)

    recorder.clear()

    assert not recorder.was_called()



from pathlib import Path

import pytest
from docutils.core import publish_doctree

ROOT_DIR = Path(__file__).parents[1]


def is_code_block(node):
    return node.tagname == "literal_block" and "code" in node.attributes["classes"]


def get_code_blocks(file_path):
    with file_path.open() as file_handle:
        content = file_handle.read()

    code_blocks = publish_doctree(content).findall(condition=is_code_block)
    return [block.astext() for block in code_blocks]


@pytest.mark.parametrize("rst_file", ("docs/index.rst", "README.rst"))
def test_documentation(testdir, rst_file):
    imports = "\n".join(
        [
            "import asyncio",
            "import os",
            "import sys",
            "",
            "import pytest",
            "import pytest_subprocess",
            "import subprocess",
        ]
    )

    setup_fixture = (
        "\n\n"
        "@pytest.fixture(autouse=True)\n"
        "def setup():\n"
        "    os.chdir(os.path.dirname(__file__))\n\n"
    )

    event_loop_fixture = (
        "\n\n"
        "@pytest.fixture(autouse=True)\n"
        "def event_loop(request):\n"
        "    policy = asyncio.get_event_loop_policy()\n"
        '    if sys.platform == "win32":\n'
        "        loop = asyncio.ProactorEventLoop()\n"
        "    else:\n"
        "        loop = policy.get_event_loop()\n"
        "    yield loop\n"
        "    loop.close()\n"
    )

    code_blocks = "\n".join(get_code_blocks(ROOT_DIR / rst_file))
    testdir.makepyfile(
        imports + setup_fixture + event_loop_fixture + "\n" + code_blocks
    )

    result = testdir.inline_run()
    assert result.ret == 0
